@@ -6,12 +6,12 @@
  *	in-situ mailbox rewrite, no tempfiles, short locking,		*
  *	standard BSD mailbox format, SysV Content-Length extension	*
  *									*
- *	Copyright (c) 1996, S.R. van den Berg, The Netherlands		*
+ *	Copyright (c) 1996-1997, S.R. van den Berg, The Netherlands	*
  *	#include "README"						*
  ************************************************************************/
 #ifdef RCS
 static /*const*/char rcsid[]=
- "$Id: cucipop.c,v 1.8 1996/12/26 17:23:39 srb Exp $";
+ "$Id: cucipop.c,v 1.13 1997/08/10 10:08:57 srb Exp $";
 #endif
 
 #include "patchlevel.h"
@@ -38,6 +38,7 @@ static /*const*/char rcsid[]=
 #include <time.h>
 
 #include "cucipop.h"
+#include "atotime.h"
 #include "authenticate.h"
 #include "locking.h"
 
@@ -55,19 +56,22 @@ static /*const*/char rcsid[]=
 
 static const char*const nullp,From_[]=FROM,XFrom_[]=XFROM,cucipopn[]="cucipop",
  Received[]=
-  "Received: from %s for %s\r\n with Cubic Circle's %s (%s) %.24s\r\n",
+  "Received: by %s for %s\r\n (with Cubic Circle's %s (%s) %.24s)\r\n",
  version[]=VERSION;
 static char linebuf[LINEBUFLEN+1],*arg1,*arg2,user[ARGLEN+1],*host,*timestamp;
-static FILE*fmbox,*sockin,*sockout;
+static FILE*fmbox,*sockout;
+static size_t hostnduser;
+FILE*sockin;
+long spoofcheck;
 static struct sockaddr_in peername;
 
 struct msg
  { off_t start;
 #ifdef UIDL
-   unsigned long uidl;
+   uidl_t uidl;
 #endif
    long virtsize;
-   unsigned char deleted;
+   unsigned char deleted,tooold;
  }*msgs;
 
 static size_t msgs_filled,msgs_max,totmsgs,transfmsgs;
@@ -100,8 +104,12 @@ static size_t findmsg(void)
 }
 
 static const char*bintohex(const void*ptr,size_t size)
-{ static char hex[BIN2HEXLEN+1],*hp;
-  for(hp=hex;size--;ptr=(char*)ptr+1,hp+=2)
+{ static char hex[BIN2HEXLEN+1],*hp;int inc=1;
+  { const int test=1;
+    if(!*(const char*)&test)
+       ptr=(const char*)ptr+size+(inc=-1);
+  }
+  for(hp=hex;size--;ptr=(const char*)ptr+inc,hp+=2)
      sprintf(hp,"%02x",*(unsigned char*)ptr);
   return hex;
 }
@@ -209,15 +217,16 @@ static const auth_identity*checkchallenge(const char*ch)
   if(pass=auth_finduser(user,fileno(sockin)))
    { size_t seclen,stlen;
      ;{ char*secret;
-	if(!(secret=auth_getsecret(pass))||!*secret)
+	if(!(secret=(char*)auth_getsecret(pass))||!*secret)
 	   goto ret0;
 	if(!(timestamp=realloc(timestamp,
-	 (seclen=strlen(secret))+(stlen=strlen(timestamp))+1)))
+	 (stlen=strlen(timestamp))+(seclen=strlen(secret))+1)))
 	   outofmem();
 	strcpy(timestamp+stlen,secret);wipestring(secret);
       }
      ;{ MD5_CTX context;unsigned char digest[MD5_DIGLEN];
-	MD5Init(&context);MD5Update(&context,(unsigned char*)timestamp,seclen);
+	MD5Init(&context);
+	MD5Update(&context,(unsigned char*)timestamp,stlen+seclen);
 	wipestring(timestamp+stlen);MD5Final(digest,&context);
 	if(!strcmp(Bintohex(digest),ch))
 	   return setuid(auth_whatuid(pass))?(auth_identity*)0:pass;
@@ -241,10 +250,10 @@ static void addblock(const off_t start)
   msgs[msgs_filled++].deleted=0;
 }
 
-static int openmailbox(const auth_identity*pass)
-{ off_t start,end;size_t virtsize,extra;unsigned long uidl;unsigned inheader;
-  long clen;int fdsockin;long spoofcheck=0;
-  alarm(0);clen=end=start=0;uidl=tsabotage;fdsockin=fileno(sockin);
+static int openmailbox(const auth_identity*pass,const time_t agecutoff)
+{ off_t start,end;size_t virtsize,extra;uidl_t uidl;unsigned inheader;
+  long clen;int fdsockin;
+  alarm(0);clen=end=start=0;fdsockin=fileno(sockin);
   if(!(fmbox=fopen(mailbox=auth_mailboxname((auth_identity*)pass),"r+b")))
    { if(errno==ENOENT)
 	goto nomailbox;					 /* no mailbox found */
@@ -260,8 +269,8 @@ static int openmailbox(const auth_identity*pass)
       }
      oldend=end=stbuf.st_size;
    }
-  virtsize=extra=STRLEN(XFrom_)-STRLEN(From_)+STRLEN(Received)+strlen(host)-2+
-   strlen(user)-2+STRLEN(cucipopn)-2+STRLEN(version)-2+24-5;
+  virtsize=extra=STRLEN(XFrom_)-STRLEN(From_)+STRLEN(Received)+hostnduser-2*2+
+   STRLEN(cucipopn)-2+STRLEN(version)-2+24-5;
   fcntl(fdsockin,F_SETFL,(long)O_NONBLOCK);	/* non-blocking spoof enable */
   for(;;)
    { unsigned newlines=1;
@@ -278,6 +287,7 @@ static int openmailbox(const auth_identity*pass)
 	   newlines=2;
 	   if(!*++p)
 	    { inheader=1;addblock(start);msgs[msgs_filled-1].deleted=0;
+	      msgs[msgs_filled-1].tooold=1;
 	      if(msgs_filled>1)
 	       { totsize+=msgs[msgs_filled-2].virtsize=
 		  virtsize+start-msgs[msgs_filled-2].start;
@@ -286,6 +296,27 @@ static int openmailbox(const auth_identity*pass)
 #endif
 	       }
 	      uidl=tsabotage;virtsize=extra;
+#define DATELEN 20
+	      if(end>DATELEN+4)
+	       { char circ[DATELEN],circo[DATELEN+1];unsigned rot,frot;
+		 for(circo[DATELEN]=frot=rot=0;;rot=(rot+1)%DATELEN)
+		  { if(!end)
+		       goto nomailbox;
+		    end--;virtsize++;start++;
+		    if((c=getc(fmbox))=='\n')
+		       break;
+		    circ[rot]=c;frot++;
+		  }
+		 if(frot>=DATELEN)
+		  { for(frot=0;frot<DATELEN;frot++,rot=(rot+1)%DATELEN)
+		       circo[frot]=circ[rot];
+		    if(atotime(circo)-agecutoff>=0)
+		     { if(tsabotage)
+			  uidl=0;
+		       msgs[msgs_filled-1].tooold=0;
+		     }
+		  }
+	       }
 	      break;
 	    }
 	 }
@@ -392,10 +423,11 @@ static int rwrite(const int fd,char*a,const unsigned len)
 
 #ifdef UIDL
 static void printuidl(size_t i)
-{ (sizeof msgs[i].uidl+sizeof msgs[i].virtsize)*2>UIDLLEN?
+{ size_t sz=msgs[i].virtsize-hostnduser;
+  (sizeof msgs[i].uidl+sizeof msgs[i].virtsize)*2>UIDLLEN?
    fail("uidl too long"),terminate(EX_SOFTWARE),0:0;
   fprintf(sockout,"%ld %s",(long)(i+1),Bintohex(msgs[i].uidl));
-  fprintf(sockout,"%s\r\n",Bintohex(msgs[i].virtsize));
+  fprintf(sockout,"%s\r\n",Bintohex(sz));
 }
 #endif
 
@@ -406,8 +438,8 @@ void printusg(void)
 main(int argc,const char*const argv[])
 { unsigned loginattempts=0,quiet=0,debug=0;short port=POP3_PORT;
   unsigned douser=1,douidl=1,doapop=1,dotop=1,autodel=0,sabotage=0;
-  static const char tdotnl[]=".\r\n";
-  chdir(ROOTDIR);openlog(cucipopn,LOG_PID,LOG_MAIL);sockin=stdin;
+  static const char tdotnl[]=".\r\n";time_t agecutoff=-AGETOLERANCE;
+  chdir(ROOTDIR);openlog(cucipopn,LOG_PID,LOG_FACILITY);sockin=stdin;
   sockout=stdout;
   if(!(msgs=malloc(GROWSTEP*sizeof(struct msg))))
      outofmem();
@@ -475,13 +507,16 @@ main(int argc,const char*const argv[])
 #ifdef UIDL
 		  " UIDL"
 #endif
-		  "\n",cucipopn,version);
+		  "\nSystem mailbox:\t\t%s\n",cucipopn,version,
+		  auth_mailboxname(auth_finduid(getuid(),0)));
 		 return EXIT_SUCCESS;
 usg:	      default:printusg();
 		 return EX_USAGE;
 	      case HELP1:case HELP2:printusg();
 		 fprintf(stderr,XUSAGE,POP3_PORT);
 		 return EX_USAGE;
+	      case EXPIRE:Qnext_arg();agecutoff=atosec(chp,&chp);
+		 continue;
 	      case PORT:Qnext_arg();port=strtol(chp,(char**)&chp,10);
 		 continue;
 	      case '\0':;
@@ -540,7 +575,7 @@ usg:	      default:printusg();
      uname(&name);
      ;{ size_t hlen;
 	if(!(host=malloc(hlen=strlen(name.nodename)+1))||
-	 !(timestamp=malloc(1+sizeof(pid_t)*2+sizeof(time_t)*2+hlen+1)))
+	 !(timestamp=malloc(1+sizeof(pid_t)*2+sizeof(time_t)*2+1+hlen+1)))
 	   outofmem();
       }
      ;{ pid_t pid=getpid();
@@ -549,6 +584,7 @@ usg:	      default:printusg();
      ;{ time_t t=time((time_t*)0);
 	if(sabotage)
 	   tsabotage=t;
+	agecutoff=t-agecutoff;
 	sprintf(timestamp+1+sizeof(pid_t)*2,"%s@%s>",Bintohex(t),
 	 strcpy(host,name.nodename));
       }
@@ -599,7 +635,8 @@ usg:	      default:printusg();
 	    { strcpy(user,arg1);
 	      if(pass=checkchallenge(arg2))
 okpass:	       { free(timestamp);
-		 if(openmailbox(pass))
+		 hostnduser=strlen(host)+strlen(user);
+		 if(openmailbox(pass,agecutoff))
 		    fprintf(sockout,"+OK Congratulations!\r\n");
 		 else
 		  { syslog(LOG_ALERT,"Error opening %s's mailbox",user);
@@ -720,7 +757,7 @@ prequitstate:		{ autodel=2;
 	      putc(c,sockout);
 	    }
 	   fprintf(sockout,tdotnl);
-	   if(dodel&&!fflush(sockout))
+	   if(dodel&&!fflush(sockout)&&msgs[dodel-1].tooold)
 	      msgs[dodel-1].deleted|=2;
 	 }
 	else
@@ -789,21 +826,21 @@ tdot:	   fprintf(sockout,tdotnl);
 	 "NOOP, RSET, QUIT");
    }
 quitstate:
-  ;{ off_t target=0;
-     if(msgs_filled)
-      { char*buf;off_t source;size_t slen,i;int fd=fileno(fmbox);
+  ;{ off_t target=0;size_t i;
+     for(i=0;i<msgs_filled&&!msgs[i].deleted;i++);
+     if(i<msgs_filled)
+      { char*buf;off_t source;size_t slen;int fd=fileno(fmbox);
 	alarm(0);setbuf(sockin,(char*)0);setbuf(fmbox,(char*)0);free(host);
+	if(!(buf=malloc(COPYBUF)))
+	   outofmem();
 	if(!lockmbox(fd,mailbox))
 	 { syslog(LOG_ALERT,"Error locking %s's mailbox",user);
 	   fail("Error locking your mailbox");terminate(EX_NOINPUT);
 	 }
-	if(!(buf=malloc(COPYBUF)))
-	   outofmem();
 	msgs[msgs_filled++].deleted=0;
 	if(oldend>(source=lseek(fd,(off_t)0L,SEEK_END)))
 	   goto corrupt;
 	addblock(source);msgs[--msgs_filled].deleted=1;
-	for(i=0;i<msgs_filled&&!msgs[i].deleted;i++);
 	for(target=msgs[i].start;i<msgs_filled;i++)
 	   if(!msgs[i].deleted)
 	    { size_t readin;
